import asyncio
import collections
import datetime
import typing

import aiohttp
from tabulate import tabulate

from fwtv import verifier

LIST_JSON_RESPONSE = typing.List[typing.Dict[str, typing.Any]]
GET_ERRORS = typing.Tuple[typing.Dict[str, typing.List[str]], typing.Dict[str, typing.List[verifier.Error]]]


class AuthenticationError(Exception):
    """
    Exception to be raised if there was an error authenticating.
    """


class FactorialApi:
    """
    Factorial api wrapper providing methods to get all employees and attendances.
    """

    def __init__(self, api_key: str, base_url="https://api.factorialhr.com"):
        self.headers = {"accept": "application/json", "x-api-key": api_key}
        self.session = aiohttp.ClientSession(base_url, headers=self.headers)

    async def close(self):
        await self.session.close()

    async def __aexit__(self, *_, **__):
        await self.close()

    async def __aenter__(self) -> "FactorialApi":
        return self

    async def _get(self, endpoint: str) -> typing.Any:
        async with self.session.get("/api/" + endpoint) as resp:
            if resp.status == 401:
                raise AuthenticationError("Authentication failed because the api key may be invalid")
            elif resp.status != 200:
                raise ValueError(f'Error getting "/api/{endpoint}": {resp.status}')
            return await resp.json()

    async def get_attendances(self) -> LIST_JSON_RESPONSE:
        return typing.cast(LIST_JSON_RESPONSE, await self._get("v2/time/attendance"))

    async def get_employees(self) -> LIST_JSON_RESPONSE:
        return typing.cast(LIST_JSON_RESPONSE, await self._get("v2/core/employees"))


def convert_timestamp(timestamp: str) -> datetime.datetime:
    """
    Converts iso timestamp to datetime

    @param timestamp: iso timestamp
    @return: converted timestamp
    """
    return datetime.datetime.fromisoformat(timestamp.replace("Z", ""))


def get_errors(
    start: datetime.datetime, end: datetime.datetime, attendances: LIST_JSON_RESPONSE, employees: LIST_JSON_RESPONSE
) -> GET_ERRORS:
    preconditions: typing.Dict[str, typing.List[str]] = collections.defaultdict(list)
    errors: typing.Dict[str, typing.List[verifier.Error]] = collections.defaultdict(list)
    for employee in employees:
        name = employee["full_name"]
        employee_attendances: typing.List[verifier.Attendance] = []

        for attendance in filter(lambda x: x["employee_id"] == employee["id"], attendances):
            if "clock_in" not in attendance or not attendance["clock_in"]:
                preconditions[name].append(f'no clock in time provided for attendance with id "{attendance["id"]}"')
                continue
            clock_in = convert_timestamp(attendance["clock_in"])
            if clock_in < start or clock_in > end:
                continue
            if "clock_out" not in attendance or not attendance["clock_out"]:
                preconditions[name].append(f'no clock out time provided for clock in time "{clock_in}"')
                continue
            if not attendance["workable"]:
                continue  # it has been declared as a break
            try:
                a = verifier.Attendance(clock_in=clock_in, clock_out=convert_timestamp(attendance["clock_out"]))
                employee_attendances.append(a)
            except ValueError as e:
                preconditions[name].append(str(e))
                continue
        errors_ = verifier.verify_attendances(employee_attendances)
        if errors_:
            errors[name] = errors_
    return preconditions, errors


def main(start: datetime.datetime, end: datetime.datetime, api_key: str):  # pragma: no cover
    async def fetch_data() -> typing.Tuple[LIST_JSON_RESPONSE, LIST_JSON_RESPONSE]:
        async with FactorialApi(api_key) as api:
            _attendances = await api.get_attendances()
            _employees = await api.get_employees()
        return _employees, _attendances

    employees, attendances = asyncio.run(fetch_data())

    preconditions, errors = get_errors(start, end, attendances, employees)
    entries = []
    for name, precondition_errors in preconditions.items():
        for precondition_error in precondition_errors:
            entries.append([name, precondition_error])
    print("Precondition errors")
    print(tabulate(entries, headers=["Name", "Error"], tablefmt="orgtbl"))
    print("")

    headers = ["Name", "Day(s)", "Reason", "Break", "Attended"]
    entries = []
    for name, error in errors.items():
        for i, e in enumerate(error):
            affected_days = [str(day) for day in e.days_affected]
            affected_days.sort()
            entries.append([name, ", ".join(affected_days), e.reason, str(e.break_time), str(e.time_attended)])
            if i + 1 == len(error):
                entries.append(["", "", "", "", ""])
    print("Working time verification failures")
    print(tabulate(entries[:-1], headers=headers, tablefmt="orgtbl"))


def cli():  # pragma: no cover
    import argparse

    parser = argparse.ArgumentParser(
        "Fetch attendances and employees from FactorialHR api and verify that the " "work time comply with german rules"
    )
    parser.add_argument("start", type=datetime.datetime.fromisoformat, help="Start iso 8601 formatted timestamp")
    parser.add_argument("end", type=datetime.datetime.fromisoformat, help="End iso 8601 formatted timestamp")
    parser.add_argument("api_key", type=str, help="Company api key")
    args = parser.parse_args()
    main(args.start, args.end, args.api_key)



import dataclasses
import datetime
import typing

HOURS_6 = datetime.timedelta(hours=6)
HOURS_9 = datetime.timedelta(hours=9)
HOURS_10 = datetime.timedelta(hours=10)
MINUTES_30 = datetime.timedelta(minutes=30)
MINUTES_45 = datetime.timedelta(minutes=45)


@dataclasses.dataclass(frozen=True)
class Attendance:
    """
    A timespan relevant for verification.
    """

    clock_in: datetime.datetime
    clock_out: datetime.datetime

    def __post_init__(self):
        if self.clock_out <= self.clock_in:
            raise ValueError(f"Clocked out earlier or at the same time as clocked in for attendance {self}")

    def __str__(self) -> str:
        return f"{type(self).__name__}(clock_in={str(self.clock_in)}, clock_out={self.clock_out})"

    def __repr__(self) -> str:
        return str(self)


def calculate_time_attended(attendances: typing.List[Attendance]) -> datetime.timedelta:
    """
    Calculates the time attended.

    @param attendances: list of attendances
    @return: time attended
    """
    attendances.sort(key=lambda x: x.clock_in)
    attended = datetime.timedelta()
    clock_out = None
    for i, attendance in enumerate(attendances):
        # check overlaying attendances
        if not clock_out or clock_out < attendance.clock_in:
            clock_in = attendance.clock_in
        else:
            clock_in = clock_out
        if not clock_out or clock_out < attendance.clock_out:
            clock_out = attendance.clock_out
            attended += clock_out - clock_in
    return attended


def calculate_break_time(attendances: typing.List[Attendance]) -> datetime.timedelta:
    """
    Calculates the time between the specified attendances.

    @param attendances: list of attendances
    @return: time between attendances
    """
    if not attendances:
        return datetime.timedelta(seconds=0)
    attendances.sort(key=lambda x: x.clock_in)
    clock_in = attendances[0].clock_in
    attendances.sort(key=lambda x: x.clock_out)
    clock_out = attendances[-1].clock_out
    return clock_out - clock_in - calculate_time_attended(attendances)


@dataclasses.dataclass(frozen=True)
class Error:
    """
    Error found during verification.
    """

    reason: str
    attendances: typing.List[Attendance]

    @property
    def days_affected(self) -> typing.Set[datetime.date]:
        days: typing.Set[datetime.date] = set()
        for attendance in self.attendances:
            days = days.union({attendance.clock_in.date(), attendance.clock_out.date()})
        return days

    @property
    def break_time(self) -> datetime.timedelta:
        return calculate_break_time(self.attendances)

    @property
    def time_attended(self) -> datetime.timedelta:
        return calculate_time_attended(self.attendances)


def verify_attendances(attendances: typing.List[Attendance]) -> typing.List[Error]:
    """
    Verifies that the specified attendances meet the following requirements (in order):
      1. It shall not be allowed to attend for more than 10 hours without not attended for at least 11 hours
      2. It shall not be allowed to attend for more than 9 hours without not attended for at least 45 minutes
      3. It shall not be allowed to attend for more than 6 hours without not attended for at least 30 minutes

    @param attendances: attendances to be verified
    @return: a list of errors found during verification
    """
    errors: typing.List[Error] = []
    attendances.sort(key=lambda x: (x.clock_out, x.clock_out))  # ensure correct order
    # contains all attendances currently in verification
    # will be reset once a break of 11 hours has been reached
    current_attendances: typing.List[Attendance] = []
    for i, attendance in enumerate(attendances):
        current_attendances.sort(key=lambda x: x.clock_out)
        if current_attendances and attendance.clock_in > current_attendances[-1].clock_out:
            break_time = attendance.clock_in - current_attendances[-1].clock_out
        else:
            break_time = datetime.timedelta(seconds=0)  # first attendance, there is no break
        if break_time > datetime.timedelta(hours=11):
            # reset relevant as 11-hour break has reached between previous attendances and current attendance
            current_attendances = [attendance]
        else:
            current_attendances.append(attendance)

        cumulated_time_attended = calculate_time_attended(current_attendances)
        cumulated_break_time = calculate_break_time(current_attendances)
        reason = None
        if cumulated_time_attended > HOURS_6 and cumulated_break_time < MINUTES_30:
            reason = "Attended more than 6 hours without a cumulated break of 30 min"
        if cumulated_time_attended > HOURS_9 and cumulated_break_time < MINUTES_45:
            reason = "Attended more than 9 hours without a cumulated break of 45 min"
        if cumulated_time_attended > HOURS_10:
            reason = "Attended more than 10 hours without a single break of 11 hours"
        if reason:
            errors.append(Error(reason=reason, attendances=current_attendances[:]))

    return errors
